/*
    ext2_journal.c -- ext2 resize journalling support
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static const char _ext2_journal_c[] = "$Id: ext2_journal.c,v 1.10 1999/07/27 12:30:57 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include "ext2.h"

struct ext2_journal
{
	ino_t parent;
	ino_t inode;
	blk_t size;
};


static blk_t ext2_find_free_block(struct ext2_fs *fs, int group)
{
	blk_t offset;
	int   grp;
	blk_t i;

	grp = group;

	while (!fs->gd[grp].bg_free_blocks_count)
	{
tryagain:
		grp = (grp + 1) % fs->numgroups;

		if (grp == group)
		{
			fprintf(stderr, "filesystem full!\n");
			exit(-1);
		}
	}

	offset = grp * fs->sb.s_blocks_per_group + fs->sb.s_first_data_block;

	for (i=0;i<fs->sb.s_blocks_per_group;i++)
		if (ext2_is_data_block(fs, i) && !ext2_get_block_state(fs, offset + i))
			return offset + i;

	fprintf(stderr, "serious error in ext2_find_free_block!\n");
	goto tryagain;
}

static ino_t ext2_find_free_inode(struct ext2_fs *fs)
{
	ino_t i;
	int   grp;
	ino_t offset;

	grp = 0;

	while (!fs->gd[grp].bg_free_inodes_count)
	{
tryagain:
		grp++;

		if (grp == fs->numgroups - 1)
		{
			fprintf(stderr, "filesystem out of inodes!\n");
			exit(-1);
		}
	}

	offset = grp * fs->sb.s_inodes_per_group + 1;
	for (i=0;i<fs->sb.s_inodes_per_group;i++)
		if (!ext2_get_inode_state(fs, offset + i))
			return offset + i;

	fprintf(stderr, "serious error in ext2_find_free_inode()\n");
	goto tryagain;
}

static ino_t find_file_in_dir_block(struct ext2_fs *fs, blk_t block, char *name)
{
	struct ext2_buffer_head *bh;
	ino_t inode;
	off_t offset;

	bh = ext2_bread(fs, block);
	inode = 0;
	offset = 0;

	do
	{
		struct ext2_dir_entry *ptr;

		ptr = (struct ext2_dir_entry *)(bh->data + offset);
		if (ptr->name_len == strlen(name) && !memcmp(ptr->name, name, strlen(name)))
		{
			inode = ptr->inode;
			goto out;
		}

		offset += ptr->rec_len;
	} while (offset < fs->blocksize);

 out:
	ext2_brelse(bh, 0);
	return inode;
}


static ino_t find_lost_and_found_inode(struct ext2_fs *fs)
{
	int i;
	struct ext2_inode inode;

	ext2_read_inode(fs, (ino_t)2, &inode);
	for (i=0;i<12;i++)
	{
		ino_t inum;

		if ((inum = find_file_in_dir_block(fs, inode.i_block[i], "lost+found")) != 0)
			return inum;
	}

	return 0;
}

static ino_t create_journal_file(struct ext2_fs *fs)
{
	int		  i;
	struct ext2_inode inode;
	ino_t		  inodenum;
	blk_t		  numblocks;

	numblocks = ext2_journal_size >> (fs->logsize - 10);

	if (!fs->sb.s_free_inodes_count)
	{
		fprintf(stderr, "need at least 1 free inode for a journal!\n");
		return 0;
	}

	if (fs->sb.s_free_blocks_count < numblocks+1)
	{
		fprintf(stderr, "need at least %i free data blocks\n", numblocks+1);
		return 0;
	}

	inodenum = ext2_find_free_inode(fs);

	inode.i_mode = 0100400;
	inode.i_size = ext2_journal_size << 10;
	inode.i_uid =
		inode.i_gid = 0;
	inode.i_atime =
		inode.i_ctime =
		inode.i_mtime =
		inode.i_dtime = 0;
	inode.i_links_count = 1;
	inode.i_blocks = (ext2_journal_size + (numblocks>12?1:0)) << 1;
	inode.i_flags = 0;
	inode.osd1.linux1.l_i_reserved1 = 0;
	inode.i_block[0] = 0;
	inode.i_block[1] = 0;
	inode.i_block[2] = 0;
	inode.i_block[3] = 0;
	inode.i_block[4] = 0;
	inode.i_block[5] = 0;
	inode.i_block[6] = 0;
	inode.i_block[7] = 0;
	inode.i_block[8] = 0;
	inode.i_block[9] = 0;
	inode.i_block[10] = 0;
	inode.i_block[11] = 0;
	inode.i_block[12] = 0;
	inode.i_block[13] = 0;
	inode.i_block[14] = 0;
	inode.i_version = 0;
	inode.i_file_acl = 0;
	inode.i_dir_acl = 0;
	inode.i_faddr = 0;
	inode.osd2.linux2.l_i_frag = 0;
	inode.osd2.linux2.l_i_fsize = 0;
	inode.osd2.linux2.i_pad1 = 0;
	inode.osd2.linux2.l_i_reserved2[0] = 0;
	inode.osd2.linux2.l_i_reserved2[1] = 0;

	if (numblocks > 12)
	{
		blk_t block;

		block = ext2_find_free_block(fs, 0);
		ext2_set_block_state(fs, block, 1, 1);
		ext2_zero_blocks(fs, block, 1);
		inode.i_block[12] = block;
	}

	for (i=0;i<numblocks;i++)
	{
		blk_t block;

		block = ext2_find_free_block(fs, 0);
		ext2_set_block_state(fs, block, 1, 1);
		ext2_zero_blocks(fs, block, 1);

		if (i >= 12)
		{
			struct ext2_buffer_head *bh;

			bh = ext2_bread(fs, inode.i_block[12]);
			bh->dirty = 1;
			((__u32 *)bh->data)[i-12] = block;
			ext2_brelse(bh, 0);
		}
		else
			inode.i_block[i] = block;
	}

	ext2_write_inode(fs, inodenum, &inode);
	ext2_set_inode_state(fs, inodenum, 1, 1);

	return inodenum;
}

static ino_t find_journal_file(struct ext2_fs *fs, ino_t parent)
{
	int i;
	struct ext2_inode inode;
	ino_t inodenum;

	ext2_read_inode(fs, parent, &inode);

	for (i=0;i<12;i++)
		if ((inodenum = find_file_in_dir_block(fs, inode.i_block[i], "journal")) != 0)
			return inodenum;

	return 0;
}

/* FIXXXME this is currently non-optimal (it will always start at a new block) */
static int add_file_to_dir_block(struct ext2_fs *fs, blk_t block, ino_t file)
{
	struct ext2_buffer_head *bh;
	off_t offset;

	bh = ext2_bread(fs, block);

	offset = 0;
	do
	{
		struct ext2_dir_entry *ptr;

		ptr = (struct ext2_dir_entry *)(bh->data + offset);

		if (!ptr->name_len)
		{
			ptr->inode = file;
			ptr->rec_len = 16;
			ptr->name_len = 7;
			strcpy(ptr->name, "journal");

			ptr = (struct ext2_dir_entry *)(bh->data + offset + 16);
			ptr->rec_len = fs->blocksize - offset - 16;
			ptr->name_len = 0;

			bh->dirty = 1;
			ext2_brelse(bh, 1);
			return 1;
		}

		offset += ptr->rec_len;
	} while (offset < fs->blocksize);

	ext2_brelse(bh, 1);
	return 0;
}

static void add_file_to_dir(struct ext2_fs *fs, ino_t directory, ino_t file)
{
	int i;
	struct ext2_inode inode;

	ext2_read_inode(fs, directory, &inode);

	for (i=0;i<12;i++)
		if (add_file_to_dir_block(fs, inode.i_block[i], file))
			return;

	fprintf(stderr, "/lost+found is full?");
	exit(-1);
}

int ext2_journal_init(struct ext2_fs *fs)
{
	ino_t lostnfound;
	ino_t journal;

	return 1;       /* FIXXXME */

	/* FIXXXME */
	if (ext2_journal_size > 256)
	{
		fprintf(stderr, "max journal size is currently 256\n");
		ext2_journal_size = 256;
	}

	if ((lostnfound = find_lost_and_found_inode(fs)) == 0)
	{
		fprintf(stderr, "error finding /lost+found inode!\n");
		return 0;
	}

	if ((journal = find_journal_file(fs, lostnfound)) != 0)
	{
		fprintf(stderr, "we have a journal already!\n");
		fprintf(stderr, "replay it now\n");
		fprintf(stderr, "delete it now\n");
		return 1;
	}

	fprintf(stderr, "creating journal\n");

	if ((journal = create_journal_file(fs)) == 0)
	{
		fprintf(stderr, "error creating journal file!\n");
		return 0;
	}

	fprintf(stderr, "journal is %li\n", journal);

	add_file_to_dir(fs, lostnfound, journal);

	return 1;
}

void ext2_journal_deinit(struct ext2_fs *fs)
{
}
